@@ -4,9 +4,17 @@ import logging
 
 import networkx as nx
 import numpy as np
-import ray
+from joblib import Parallel, delayed
 from sklearn.base import BaseEstimator
 from sklearn.linear_model import LogisticRegression
+from sklearn.utils.validation import _check_sample_weight
+
+try:
+    import ray
+except ImportError:
+    _has_ray = False
+else:
+    _has_ray = True
 
 
 def make_leveled(y):
@@ -78,6 +86,7 @@ class HierarchicalClassifier(abc.ABC):
             a single unique class.
         n_jobs : int, default=1
             The number of jobs to run in parallel. Only :code:`fit` is parallelized.
+            If :code:`Ray` is installed it is used, otherwise it defaults to :code:`Joblib`.
         classifier_abbreviation : str, default=""
             The abbreviation of the local hierarchical classifier to be displayed during logging.
         """
@@ -88,7 +97,7 @@ class HierarchicalClassifier(abc.ABC):
         self.n_jobs = n_jobs
         self.classifier_abbreviation = classifier_abbreviation
 
-    def fit(self, X, y):
+    def fit(self, X, y, sample_weight=None):
         """
         Fit a local hierarchical classifier.
 
@@ -102,6 +111,9 @@ class HierarchicalClassifier(abc.ABC):
             converted into a sparse ``csc_matrix``.
         y : array-like of shape (n_samples, n_levels)
             The target values, i.e., hierarchical class labels for classification.
+        sample_weight : array-like of shape (n_samples,), default=None
+            Array of weights that are assigned to individual samples.
+            If not provided, then each sample is given unit weight.
 
         Returns
         -------
@@ -114,7 +126,7 @@ class HierarchicalClassifier(abc.ABC):
         # Delete unnecessary variables
         self._clean_up()
 
-    def _pre_fit(self, X, y):
+    def _pre_fit(self, X, y, sample_weight):
         # Check that X and y have correct shape
         # and convert them to np.ndarray if need be
 
@@ -122,6 +134,11 @@ class HierarchicalClassifier(abc.ABC):
             X, y, multi_output=True, accept_sparse="csr"
         )
 
+        if sample_weight is not None:
+            self.sample_weight_ = _check_sample_weight(sample_weight, X)
+        else:
+            self.sample_weight_ = None
+
         self.y_ = make_leveled(self.y_)
 
         # Create and configure logger
@@ -156,19 +173,20 @@ class HierarchicalClassifier(abc.ABC):
         self.logger_.setLevel(self.verbose)
 
         # Create console handler and set verbose level
-        ch = logging.StreamHandler()
-        ch.setLevel(self.verbose)
+        if not self.logger_.hasHandlers():
+            ch = logging.StreamHandler()
+            ch.setLevel(self.verbose)
 
-        # Create formatter
-        formatter = logging.Formatter(
-            "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
-        )
+            # Create formatter
+            formatter = logging.Formatter(
+                "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
+            )
 
-        # Add formatter to ch
-        ch.setFormatter(formatter)
+            # Add formatter to ch
+            ch.setFormatter(formatter)
 
-        # Add ch to logger
-        self.logger_.addHandler(ch)
+            # Add ch to logger
+            self.logger_.addHandler(ch)
 
     def _disambiguate(self):
         self.separator_ = "::HiClass::Separator::"
@@ -288,15 +306,25 @@ class HierarchicalClassifier(abc.ABC):
                 for j in range(1, y.shape[1]):
                     y[i, j] = y[i, j].split(self.separator_)[-1]
 
-    def _fit_node_classifier(self, nodes, local_mode):
+    def _fit_node_classifier(
+        self, nodes, local_mode: bool = False, use_joblib: bool = False
+    ):
         if self.n_jobs > 1:
-            ray.init(
-                num_cpus=self.n_jobs, local_mode=local_mode, ignore_reinit_error=True
-            )
-            lcppn = ray.put(self)
-            _parallel_fit = ray.remote(self._fit_classifier)
-            results = [_parallel_fit.remote(lcppn, node) for node in nodes]
-            classifiers = ray.get(results)
+            if _has_ray and not use_joblib:
+                ray.init(
+                    num_cpus=self.n_jobs,
+                    local_mode=local_mode,
+                    ignore_reinit_error=True,
+                )
+                lcppn = ray.put(self)
+                _parallel_fit = ray.remote(self._fit_classifier)
+                results = [_parallel_fit.remote(lcppn, node) for node in nodes]
+                classifiers = ray.get(results)
+            else:
+                classifiers = Parallel(n_jobs=self.n_jobs)(
+                    delayed(self._fit_classifier)(self, node) for node in nodes
+                )
+
         else:
             classifiers = [self._fit_classifier(self, node) for node in nodes]
         for classifier, node in zip(classifiers, nodes):
@@ -310,3 +338,5 @@ class HierarchicalClassifier(abc.ABC):
         self.logger_.info("Cleaning up variables that can take a lot of disk space")
         del self.X_
         del self.y_
+        if self.sample_weight_ is not None:
+            del self.sample_weight_
